"""
"Metric kit" to equip the real-number interval/sets with extension.
This is prototypical of analogous augmentations to other intervals*/sets*
"""


class RealDomainMetric:
    """
    Metric definition for real numbers.
    """

    @staticmethod
    def adder(val1, val2):
        """The trivial adder."""
        return val1 + val2

    @staticmethod
    def subtracter(val1, val2):
        """The trivial subtracter."""
        return val1 - val2

    zero = 0



"""
Interval and interval set subclassed to work with `datetime` values.

This serves also as prototype for other extension to specific 'real-like'
domains for subclassing and which methods should be overridden (and how).

Everything else is inherited from the superclasses in a way that takes care
of ensuring operations (utils, builders, algebraic operations)
will stay in the correct domain.
"""

import datetime

from clothesline.base.base_interval_set import BaseIntervalSet
from clothesline.base.base_interval import BaseInterval
from clothesline.base.base_domain_metric import BaseDomainMetric

from clothesline.generic.interval_generic_builder import IntervalGenericBuilder
from clothesline.generic.interval_generic_utils import IntervalGenericUtils
from clothesline.generic.interval_set_generic_utils import (
    IntervalSetGenericUtils,
)  # noqa: E501


class DatetimeMetric(BaseDomainMetric):
    """
    The metric on the datetime domain just needs to take care
    of the fact that differences are 'timedelta' objects.
    """

    @staticmethod
    def adder(val1, val2):
        """standard addition."""
        return val1 + val2

    @staticmethod
    def subtracter(val1, val2):
        """standard subtraction."""
        return val1 - val2

    zero = datetime.timedelta(0)


class DatetimeInterval(BaseInterval):
    """
    Domain-specific interval subclass.
    Below are the methods that one should override when creating a subclass.

    For these intervals, values are `datetime`.
    """

    metric = DatetimeMetric

    @staticmethod
    def value_encoder(val):
        """domain encoder: datetime => timestamp (floating-point number)."""
        return val.timestamp()

    @staticmethod
    def value_decoder(val):
        """domain decoder: timestamp -> datetime."""
        return datetime.datetime.fromtimestamp(val)

    serializing_class = "DatetimeInterval"
    serializing_version = 1

    @staticmethod
    def builder():
        """
        Return a builder configured to make peg pairs into
        these types of intervals.
        """
        return IntervalGenericBuilder(
            interval_class=DatetimeInterval,
            interval_set_class=None,
        )

    @staticmethod
    def utils():
        """
        Return an "utils" object configured to create special cases of
        intervals as instance of this subclass.
        """
        return IntervalGenericUtils(interval_class=DatetimeInterval)


class DatetimeIntervalSet(BaseIntervalSet):
    """
    Domain-specific interval-set subclass.
    Below are the methods that one should override when creating a subclass.

    For these interval sets, values are `datetime`.
    """

    interval_class = DatetimeInterval

    serializing_class = "DatetimeIntervalSet"
    serializing_version = 1

    @staticmethod
    def builder():
        """
        Return a builder configured to make peg pairs into
        these types of interval sets.
        """
        return IntervalGenericBuilder(
            interval_set_class=DatetimeIntervalSet,
        )

    @staticmethod
    def utils():
        """
        Return an "utils" object configured to create special cases of
        interval sets as instance of this subclass.
        """
        return IntervalSetGenericUtils(
            interval_set_class=DatetimeIntervalSet,
        )



"""
An interval on the usual numeric domain (real numbers).
"""

from clothesline.base.base_interval import BaseInterval

#
from clothesline.generic.interval_generic_builder import IntervalGenericBuilder
from clothesline.generic.interval_generic_utils import IntervalGenericUtils
from clothesline.real_domain_metric import RealDomainMetric


class RealInterval(BaseInterval):
    """
    An interval over real numbers.
    Both a metric and serializability are defined in the following,
    in a prototypical way.

    For serializability, encoder and decoder connect a value in the domain
    (here: numbers) to something that is JSON-serializable in a deterministic
    and bijective way, i.e.
    - encoder*decoder = identity_dicts
    - decoder*encoder = identity_domain

    (infinities are treated automatically. no need to care about them.)
    Serializability metadata are provided along with the codec pair: the
    situation is the same as for RealIntervalSet (i.e. these do end up in
    the serializable dicts), so pay attention not to change them out of a whim.
    """

    metric = RealDomainMetric

    @staticmethod
    def value_encoder(val):
        """The trivial encoder."""
        return val  # noqa: PLC0116, PLC0321

    @staticmethod
    def value_decoder(val):
        """The trivial decoder."""
        return val  # noqa: PLC0116, PLC0321

    serializing_class = "RealInterval"
    serializing_version = 1

    @staticmethod
    def builder():
        """
        Create and return a "builder" for these intervals.
        Other concrete subclasses of BaseInterval need to simply
        replace `RealInterval` with whatever the class name.
        """
        return IntervalGenericBuilder(
            interval_class=RealInterval,
            interval_set_class=None,
        )

    @staticmethod
    def utils():
        """
        Create an "interval utils" object for these intervals.
        Other concrete subclasses of BaseInterval need to simply
        replace `RealInterval` with whatever the class name.
        """
        return IntervalGenericUtils(interval_class=RealInterval)



"""
A single interval with a begin and and end, either open or closed at its ends.
"""

from clothesline.algebra.symbols import (
    is_symbol,
    x_equals,
    x_lt,
    x_gt,
    x_repr,
    x_subtract,
)

#
from clothesline.exceptions import (
    InvalidValueError,
    MetricNotImplementedError,
    UnserializableItemError,
)


class BaseInterval:
    """
    A single uninterrupted interval over the domain field:
    [a,b] or (a,b) or (a,b] or [a,b)
    defined by two IntervalPeg objects. It can span to infinities.

    Concrete classes must provide builder() and utils() in a standard way
    (see the real-interval case) and, if desired,
    define metric and serializability properties as well.
    """

    # What follows should be considered by subclassers:

    metric = None

    value_encoder = None
    value_decoder = None

    serializing_class = None
    serializing_version = None

    @staticmethod
    def builder():
        """Create and return a "builder" for these intervals."""

    @staticmethod
    def utils():
        """Create an "interval utils" object for these intervals."""

    # Subclassers may stop reading here.

    def __init__(self, begin, end):
        """
        begin and end are IntervalPeg instances
        """
        if x_gt(begin.value, end.value):
            raise InvalidValueError("Interval begin must come before its end")
        if x_equals(begin.value, end.value):
            if begin.included != end.included:
                raise InvalidValueError(
                    "Contradicting inclusion for point-like interval"
                )
            if not begin.included and not end.included:
                raise InvalidValueError("Empty point-like open set is invalid")
        self.begin = begin
        self.end = end

    def __eq__(self, other):
        if isinstance(other, self.__class__):  # noqa: PLR1705
            return self.begin == other.begin and self.end == other.end
        else:
            return False

    def __hash__(self):
        return hash((self.__class__, hash(self.begin), hash(self.end)))

    def __repr__(self):
        begin_name = x_repr(self.begin.value)
        begin_paren = "[" if self.begin.included else "("
        end_name = x_repr(self.end.value)
        end_paren = "]" if self.end.included else ")"
        return f"{begin_paren}{begin_name}, {end_name}{end_paren}"

    def to_dict(self):
        """
        Return a json-encodable representation of this interval.
        """
        if self.value_encoder:  # noqa: PLR1705
            return {
                "class": self.serializing_class,
                "version": self.serializing_version,
                "pegs": [
                    self.begin.to_dict(v_encoder=self.value_encoder),
                    self.end.to_dict(v_encoder=self.value_encoder),
                ],
            }
        else:
            raise UnserializableItemError

    def contains(self, value):  # noqa: PLR0911
        """
        Test whether a value belongs to the interval.

        Infinities are allowed as 'value' argument, but never belong.
        """
        if is_symbol(value):  # noqa: PLR1705
            return False
        else:
            # value is a regular number:
            if x_lt(self.begin.value, value):
                # value to the right of begin
                if x_lt(value, self.end.value):  # noqa: PLR1705
                    # value to the left of end
                    return True
                else:
                    # value to the right, or equal to, end
                    if x_equals(self.end.value, value):  # noqa: PLR1705
                        # value sits at end
                        return self.end.included
                    else:
                        # value to the right of end
                        return False
            else:
                # value to the left, or equal to begin
                if x_equals(self.begin.value, value):  # noqa: PLR1705
                    # value sits at begin
                    return self.begin.included
                else:
                    # value to the left of begin
                    return False

    def extension(self):
        """
        If a metric is defined for this interval type,
        use it to compute this interval's 'extension'.
        """
        if self.metric:  # noqa: PLR1705
            return x_subtract(
                self.end.value,
                self.begin.value,
                subtracter=self.metric.subtracter,
            )
        else:
            raise MetricNotImplementedError

    def pegs(self):
        """Return the two ends, iterably."""
        yield self.begin
        yield self.end

    def intervals(self):
        """
        Return an 'iterable' over a single element, this interval.
        This is only to enable quick-syntax for those IntervalSet
        set-wise operations whereby the second operand is a puny Interval.
        """
        yield self



"""
One of the two ends of an interval.

A single value + indication of whether included or not in the interval it
contributes to defining.

In the chain IntervalPeg -> Interval* -> IntervalSet*, the peg is the only
level that is not subclassed/enriched. It receives the context it needs
(e.g. a codec pair) in individual calls as the peg is not indended for use
by outside of this library.
"""

from clothesline.algebra.symbols import is_symbol, x_to_dict, x_from_dict
from clothesline.algebra.symbols import x_equals

#
from clothesline.exceptions import InvalidValueError


class IntervalPeg:
    """
    A boundary value + indication of whether the value itself is
    included/excluded.

    Note: greater/lesser inequalities among instances of this class make
    no sense, whereas they do among their 'value' values.
    """

    def __init__(self, value, included):
        self.value = value
        if is_symbol(self.value) and included:
            raise InvalidValueError("Infinities cannot be included in peg")
        self.included = included

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if x_equals(self.value, other.value):  # noqa: PLR1705
                return self.included == other.included
            else:
                return False
        else:
            return False

    def __hash__(self):
        return hash((self.value, self.included))

    def to_dict(self, v_encoder):
        """
        Return a json-encodable representation of this peg.
        An encoder for the regular domain values is required.
        """
        return {
            "value": x_to_dict(self.value, v_encoder=v_encoder),
            "included": self.included,
        }

    @staticmethod
    def from_dict(input_dict, v_decoder):
        """
        Parse a dict-representation of a peg and return it as a peg.
        A decoder for the regular domain values is required.
        """
        return IntervalPeg(
            x_from_dict(input_dict["value"], v_decoder=v_decoder),
            input_dict["included"],
        )



"""
Entry point for easily exposing the useful imports
"""

from clothesline.real_interval_set import RealIntervalSet  # noqa: F401
from clothesline.enriched.datetime_interval_set import (  # noqa: F401, E501
    DatetimeIntervalSet,
)


def main():
    """Entry point for an application script"""
    print("clothesline offers no useful scripts.")



"""
An example of interval/set kit that does not admit an underlying metric.

String form an ordered set, with e.g.
"a" < "aa" < "aaa" < "aab" < "ab" < "b"
and infinite new strings insertable inbetween, with no well-defined 'distance',

Useful mostly as a sanity check that a metric is not assumed anywhere.

Also, rather arbitrarily (for testing purposes), string-intervals are set
to non-serializable.

Please refer to the Datetime case for relevant comments on the structure.
"""

from clothesline.base.base_interval_set import BaseIntervalSet
from clothesline.base.base_interval import BaseInterval

from clothesline.generic.interval_generic_builder import IntervalGenericBuilder
from clothesline.generic.interval_generic_utils import IntervalGenericUtils
from clothesline.generic.interval_set_generic_utils import (
    IntervalSetGenericUtils,
)  # noqa: E501


class StringInterval(BaseInterval):
    """
    Intervals between strings.
    """

    @staticmethod
    def builder():
        """
        The builder peg-pair -> string interval.
        """
        return IntervalGenericBuilder(
            interval_class=StringInterval,
            interval_set_class=None,
        )

    @staticmethod
    def utils():
        """
        String-interval utils
        """
        return IntervalGenericUtils(interval_class=StringInterval)


class StringIntervalSet(BaseIntervalSet):
    """
    A string-interval-set.
    """

    interval_class = StringInterval

    @staticmethod
    def builder():
        """
        String-interval-set builder.
        """
        return IntervalGenericBuilder(
            interval_set_class=StringIntervalSet,
        )

    @staticmethod
    def utils():
        """
        Return an "utils" object configured to create special cases of
        interval sets as instance of this subclass.
        """
        return IntervalSetGenericUtils(
            interval_set_class=StringIntervalSet,
        )



"""
Metric specification for an interval/set.
"""


class BaseDomainMetric:
    """
    Abstract class for creating 'metric' static classes,
    which prescribe how to deal with the concept of interval "length" and
    their summations.
    """

    def __init__(self):
        raise RuntimeError("This class is static.")

    @staticmethod
    def adder(val1, val2):  # noqa: PLW0613
        """How to add two 'extensions' to get another 'extension'."""

    @staticmethod
    def subtracter(val1, val2):  # noqa: PLW0613
        """
        How to subtract two ends of an interval (peg values)
        to get its 'extension'.
        """

    """What is the 'zero extension' when computing the extension of a set."""  # noqa: PLW0105, E501
    zero = ...



"""
A class containing the "interface" for creation of intervals*, (not 'sets').
Like for the IntervalSetGenericUtils, the actual class to instantiate
is set per-instance, so this serves as a generic class template.
"""

from clothesline.interval_peg import IntervalPeg
from clothesline.algebra.symbols import PlusInf, MinusInf

from clothesline.exceptions import (
    UnparseableDictError,
    UnserializableItemError,
    UnsupportedVersionDictError,
)


class IntervalGenericUtils:
    """
    An "interval utils" class. Instances of this, properly initialized
    with the name of the interval class to create, are able to produce several
    'standard' intervals such as "all", "open set" and so on.
    """

    def __init__(self, interval_class):
        """
        When creating an utils instance, which can create intervals*,
        the interval_class that must be passed is a function (a constructor)
        that makes its two arguments (peg0, peg1) into an interval of the
        desired type.
        """
        self.interval_class = interval_class
        self.value_decoder = self.interval_class.value_decoder
        self.serializing_class = self.interval_class.serializing_class
        self.serializing_version = self.interval_class.serializing_version

    def from_dict(self, input_dict):
        """
        Using the information in the interval class, convert a dict
        (supposedly generate from an interval of this same type)
        back into an interval.
        This function takes care of injecting decoders to the lower-level
        (i.e. peg-level) from_dict function invocations.
        """
        if not self.value_decoder:
            raise UnserializableItemError
        if input_dict.get("class") != self.serializing_class:
            raise UnparseableDictError
        # Here, in the future, version upgrade logic will be injected
        if input_dict.get("version", 0) > self.serializing_version:
            raise UnsupportedVersionDictError
        if input_dict.get("version") != self.serializing_version:
            raise UnparseableDictError
        return self.interval_class(
            IntervalPeg.from_dict(
                input_dict["pegs"][0],
                v_decoder=self.value_decoder,
            ),
            IntervalPeg.from_dict(
                input_dict["pegs"][1],
                v_decoder=self.value_decoder,
            ),
        )

    def open(self, value_begin, value_end):
        """
        Create an open interval with finite boundaries.
        """
        return self.interval(
            value_begin,
            False,
            value_end,
            False,
        )

    def closed(self, value_begin, value_end):
        """
        Create a closed interval with finite boundaries.
        """
        return self.interval(
            value_begin,
            True,
            value_end,
            True,
        )

    def point(self, value):
        """
        Create a zero-length degenerate [x, x] point-line 'interval'.
        """
        return self.closed(value, value)

    def low_slice(self, value_end, included=False):
        """
        Create an interval from -inf to a certain value.
        """
        return self.interval(
            MinusInf,
            False,
            value_end,
            included,
        )

    def high_slice(self, value_begin, included=False):
        """
        Create an interval from a value up to +inf.
        """
        return self.interval(
            value_begin,
            included,
            PlusInf,
            False,
        )

    def all(self):
        """
        Return the "whole of it" interval.
        """
        return self.interval(
            MinusInf,
            False,
            PlusInf,
            False,
        )

    def interval(self, value_begin, begin_included, value_end, end_included):
        """
        Directly create an interval from the values and the open/closed specs.
        """
        return self.interval_class(
            IntervalPeg(value_begin, begin_included),
            IntervalPeg(value_end, end_included),
        )



"""
Defines symbols for extension to +/- infinity and related methods.
"""

from clothesline.exceptions import IndeterminateFormError, UnparseableDictError


class PlusInf:  # noqa: PLR0903
    """
    Symbolic constant for "+infinity"
    """

    @staticmethod
    def __repr__():
        return "+inf"


class MinusInf:  # noqa: PLR0903
    """
    Symbolic constant for "-infinity"
    """

    @staticmethod
    def __repr__():
        return "-inf"


def is_symbol(value):
    """
    Return True if the value is a symbolic constant.
    """
    return value is PlusInf or value is MinusInf


def x_repr(val):
    """
    String representation
    """
    if is_symbol(val):  # noqa: PLR1705
        return val.__repr__()
    else:
        return val.__repr__()


def x_to_dict(value, v_encoder):
    """
    Return a json-encodable representation of this extended 'value'.
    """
    if is_symbol(value):  # noqa: PLR1705
        return {"symbol": value.__repr__()}
    else:
        # "ordinary value"
        return {"o_value": v_encoder(value)}


def x_from_dict(input_dict, v_decoder):
    """
    Extract a extended value from a dictionary item, leveraging the passed
    decoder.
    """
    if "symbol" in input_dict:  # noqa: PLR1705
        if input_dict["symbol"] == PlusInf.__repr__():  # noqa: PLR1705
            return PlusInf
        elif input_dict["symbol"] == MinusInf.__repr__():
            return MinusInf
        else:
            raise UnparseableDictError
    else:
        return v_decoder(input_dict["o_value"])


# Extensions of arithmetic to "domain + infinities"


def x_equals(val1, val2):
    """
    Equality, extended
    """
    if val1 is PlusInf:  # noqa: PLR1705
        return val2 is PlusInf
    elif val1 is MinusInf:
        return val2 is MinusInf
    else:
        return val1 == val2


def x_gt(val1, val2):
    """
    Greater-than, extended
    """
    if val1 is PlusInf:  # noqa: PLR1705
        return val2 is not PlusInf
    elif val1 is MinusInf:
        return False
    else:
        # val1 is a regular number:
        if val2 is PlusInf:  # noqa: PLR1705
            return False
        elif val2 is MinusInf:
            return True
        else:
            # both are numbers
            return val1 > val2


def x_lt(val1, val2):
    """
    Less-than, extended
    """
    return x_gt(val2, val1)  # noqa: PLW1114


def x_cmp(val1, val2):
    """
    a valid 'cmp' to use for sorting values-and-symbols
    """
    if x_equals(val1, val2):  # noqa: PLR1705
        return 0
    else:
        return -1 if x_lt(val1, val2) else +1


def x_ge(val1, val2):
    """
    Greater-or-equal, extended
    """
    return x_equals(val1, val2) or x_gt(val1, val2)


def x_le(val1, val2):
    """
    Less-or-equal, extended
    """
    return x_equals(val1, val2) or x_lt(val1, val2)


def x_sum(val1, val2, adder):  # noqa: PLR0911
    """
    Sum of two values.
    Will raise an error if the two values are the two (opposite) infinities
    """
    if val1 is PlusInf:  # noqa: PLR1705
        if val2 is PlusInf:  # noqa: PLR1705
            # +inf + +inf
            return PlusInf
        elif val2 is MinusInf:
            # +inf + -inf
            raise IndeterminateFormError
        else:
            # +inf + num
            return PlusInf
    elif val1 is MinusInf:
        if val2 is PlusInf:  # noqa: PLR1705, PLR1720
            # -inf + +inf
            raise IndeterminateFormError
        elif val2 is MinusInf:
            # -inf + -inf
            return MinusInf
        else:
            # -inf + num
            return MinusInf
    else:
        if val2 is PlusInf:  # noqa: PLR1705
            # num + +inf
            return PlusInf
        elif val2 is MinusInf:
            # num + -inf
            return MinusInf
        else:
            # num + num
            return adder(val1, val2)


def x_subtract(val1, val2, subtracter):  # noqa: PLR0911
    """
    Evaluate the difference val1 - val2.
    Will raise an error if indeterminate forms arise.
    """
    if val1 is PlusInf:  # noqa: PLR1705
        if val2 is PlusInf:  # noqa: PLR1705, PLR1720
            # +inf - +inf
            raise IndeterminateFormError
        elif val2 is MinusInf:
            # +inf - -inf
            return PlusInf
        else:
            # +inf - num
            return PlusInf
    elif val1 is MinusInf:
        if val2 is PlusInf:  # noqa: PLR1705
            # -inf - +inf
            return MinusInf
        elif val2 is MinusInf:
            # -inf - -inf
            raise IndeterminateFormError
        else:
            # -inf - num
            return MinusInf
    else:
        if val2 is PlusInf:  # noqa: PLR1705
            # num - +inf
            return MinusInf
        elif val2 is MinusInf:
            # num - -inf
            return PlusInf
        else:
            # num - num
            return subtracter(val1, val2)



"""
interval algebra primitives.
"""

from functools import cmp_to_key

from clothesline.algebra.symbols import x_cmp
from clothesline.interval_peg import IntervalPeg
from clothesline.exceptions import InvalidCombineEndState


def combine_intervals(  # noqa: PLR0914, PLR0912
    int_maker,
    interval_iterables,
    combiner_function=lambda q: q[0],
):
    """
    The main workhorse for interval algebra.
    A list of N iterables over intervals is combined according to some
    prescriptions.

    `int_maker` is a function from (peg1, peg2) to an instance of the desired
    interval* class (e.g. the constructor "Interval" itself).

    `interval_iterables` is a list of iterables, each consisting of several
    intervals. Each of the iterables is an "operand" in some arbitrary
    (set-, i.e. boolean-) prescription to combine them.

    With a single iterable as input (N=1), i.e.
    `interval_iterables = [[int1, int2, ...]]`
    these are made into a normalized-form list of intervals.

    With two iterables as input (N=2),
    `interval_iterables = [[intA1, intA2, ...], [intB1, intB2, ...]]`

    one can achieve
    - setA U setB (default)
    - setA - setB
    - setA ^ setB
    depending on the passed combiner_function.

    `combiner_function` is a function from a tuple of N booleans
    to a single boolean: it specifies whether a point or an open interval
    will be in the result according to whether it was contained in each of the
    N inputs:
    - (q0, q1) => q0 or q1            # for union
    - (q0, q1) => q0 and (not q1)     # for set difference
    - (q0, q1) => q0 xor q1           # for '^'

    N > 2 will presumably never be used.
    """

    interval_lists = [
        list(interval_ite) for interval_ite in interval_iterables
    ]  # noqa: E501
    n_i_lists = len(interval_lists)

    # 1. 'split' phase
    markers = sorted(
        set(
            peg.value
            for i_list in interval_lists
            for interval in i_list
            for peg in interval.pegs()
        ),
        key=cmp_to_key(x_cmp),
    )
    m_index_map = {val: index for index, val in enumerate(markers)}

    point_included = {marker: {} for marker in markers}
    range_included = {marker: {} for marker in markers}
    # looping labeling to be tweaked for multi-source case
    for i_list_index, i_list in enumerate(interval_lists):
        for interval in i_list:
            # ends pointlike status
            for peg in interval.pegs():
                if peg.included:
                    point_included[peg.value][i_list_index] = True
            # internal pointlike values, if any
            begin_peg = interval.begin
            end_peg = interval.end
            begin_m_index = m_index_map[begin_peg.value]
            end_m_index = m_index_map[end_peg.value]
            for internal_m_index in range(begin_m_index + 1, end_m_index):
                internal_marker = markers[internal_m_index]
                point_included[internal_marker][i_list_index] = True
            # ranges (first + internal if any)
            for m_index in range(begin_m_index, end_m_index):
                marker = markers[m_index]
                range_included[marker][i_list_index] = True

    # 2. 'project' phase, using combiner_function
    point_merged = {marker: False for marker in markers}
    range_merged = {marker: False for marker in markers}
    for marker in markers:
        point_merged[marker] = combiner_function(
            [
                point_included[marker].get(i_list_index, False)
                for i_list_index in range(n_i_lists)
            ]
        )
    for marker in markers[:-1]:
        range_merged[marker] = combiner_function(
            [
                range_included[marker].get(i_list_index, False)
                for i_list_index in range(n_i_lists)
            ]
        )

    # # 3. 'merge' phase
    final_intervals = []
    i_buffer = []  # a mutable state
    for m_index, marker in enumerate(markers):
        next_is_range = range_merged[marker]
        point_included = point_merged[marker]
        if not i_buffer:
            if next_is_range:
                i_buffer = [(marker, point_included), None]
            else:
                if point_included:
                    # this point, isolated, is a zero-length interval:
                    final_intervals.append(
                        int_maker(
                            IntervalPeg(marker, True),
                            IntervalPeg(marker, True),
                        )
                    )
                else:
                    # no-op
                    pass
        else:
            # i_buffer exists already
            if next_is_range:
                if point_included:
                    # write/extend current buffer's endpoint
                    i_buffer[1] = (marker, point_included)
                else:
                    # a hole: finalize/flush buffer and re-init it at once
                    i_buffer[1] = (marker, point_included)
                    final_intervals.append(
                        int_maker(
                            IntervalPeg(i_buffer[0][0], i_buffer[0][1]),
                            IntervalPeg(i_buffer[1][0], i_buffer[1][1]),
                        )
                    )
                    i_buffer = [(marker, point_included), None]
            else:
                # end of the interval being built: flush and reset buffer
                i_buffer[1] = (marker, point_included)
                final_intervals.append(
                    int_maker(
                        IntervalPeg(i_buffer[0][0], i_buffer[0][1]),
                        IntervalPeg(i_buffer[1][0], i_buffer[1][1]),
                    )
                )
                i_buffer = []
    #
    if i_buffer:
        raise InvalidCombineEndState("Inconsistent end state in merge phase")

    return final_intervals



"""
Any set over the domain, defined by a finite number of intervals.
"""

from functools import reduce

from clothesline.algebra import combine_intervals
from clothesline.algebra.symbols import x_sum

#
from clothesline.exceptions import MetricNotImplementedError


class BaseIntervalSet:
    """
    Any portion of the "continuous line" that is the domain
    (such as the reals + infinities), defined by an arbitrary (finite) number
    of intervals.

    Any concrete implementation has to provide: a `builder()` and a `utils()`
    (in a pretty standard way, see e.g. the reference "real*" implementation),
    as well as to give the name of the class representing the individual
    intervals this set is made of.
    Moreover, serializing signature data can be provided (if so desired and if
    serializability is supported by the underlying interval implementation).
    """

    interval_class = None

    serializing_class = None
    serializing_version = None

    @staticmethod
    def builder():
        """
        Create an interval set builder.
        """

    @staticmethod
    def utils():
        """
        Create an "utils" object, offering standard intervalset* creation.
        """

    def __init__(self, intervals):
        self._intervals = self._normalize(intervals)

    def _normalize(self, intervals):
        """
        An arbitrary input of intervals (overlapping, unsorted)
        is reduced to 'normal form' using the one-single-list
        form of the generic combiner.
        """
        return combine_intervals(self.interval_class, [intervals])

    def to_dict(self):
        """
        Return a json-encodable representation of this interval set.
        """
        return {
            "class": self.serializing_class,
            "version": self.serializing_version,
            "intervals": [interval.to_dict() for interval in self._intervals],
        }

    def contains(self, value):
        """
        Test whether a value belongs to the set.

        Conventionally, infinities do not belong to any interval set.
        """
        return any(interval.contains(value) for interval in self._intervals)

    def extension(self):
        """
        If a metric is defined for this interval type,
        use it to compute this interval set's (overall) 'extension'.
        """
        if self.interval_class.metric:  # noqa: PLR1705

            def c_sum(val1, val2):
                return x_sum(
                    val1,
                    val2,
                    self.interval_class.metric.adder,
                )

            return reduce(
                c_sum,
                (interval.extension() for interval in self._intervals),
                self.interval_class.metric.zero,
            )
        else:
            raise MetricNotImplementedError

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and len(self._intervals) == len(other._intervals)  # noqa: W503
            and self._intervals == other._intervals  # noqa: W503
        )

    def __hash__(self):
        return hash(
            (
                self.__class__,
                tuple((hash(interval) for interval in self._intervals)),
            )  # noqa: E501
        )

    def __repr__(self):
        if not self._intervals:  # noqa: PLR1705
            return "{}"
        else:
            return " U ".join(
                interval.__repr__() for interval in self._intervals
            )  # noqa: E501

    def __add__(self, other):
        """Alias for set-wise union."""
        return self.union(other)

    def __sub__(self, other):
        """Alias for set-wise difference."""
        return self.difference(other)

    def intervals(self):
        """Return an iterable over the intervals of this set."""
        for interval in self._intervals:
            yield interval

    def union(self, other):
        """
        Union of interval sets.
        """
        return self.__class__(
            combine_intervals(
                self.interval_class,
                [self._intervals, other.intervals()],
                combiner_function=lambda q: q[0] or q[1],
            )
        )

    def difference(self, other):
        """
        Difference of interval sets.
        """
        return self.__class__(
            combine_intervals(
                self.interval_class,
                [self._intervals, other.intervals()],
                combiner_function=lambda q: q[0] and not q[1],
            )
        )

    def intersect(self, other):
        """
        Intersection of interval sets.
        """
        return self.__class__(
            combine_intervals(
                self.interval_class,
                [self._intervals, other.intervals()],
                combiner_function=lambda q: q[0] and q[1],
            )
        )

    def xor(self, other):
        """
        XOR ("exclusive disjunction") of interval sets.
        """
        return self.__class__(
            combine_intervals(
                self.interval_class,
                [self._intervals, other.intervals()],
                combiner_function=lambda q: q[0] ^ q[1],
            )
        )

    def complement(self):
        """
        Set complement of the interval set.
        """
        return self.utils().all().difference(self)

    def superset_of(self, other):
        """Test whether another interval(set) is contained in this."""
        return other - self == self.utils().empty()



"""
An arbitrary set defined by a finite number of intervals.
This is the base form, which has no metric imposed and assumes trivial
serializability of its values.
"""

from clothesline.base.base_interval_set import BaseIntervalSet
from clothesline.generic.interval_generic_builder import IntervalGenericBuilder
from clothesline.generic.interval_set_generic_utils import (
    IntervalSetGenericUtils,
)  # noqa: E501

#
from clothesline.real_interval import RealInterval


class RealIntervalSet(BaseIntervalSet):
    """
    A set over the real numbers, such as:
        (-inf, -5] U (-3, 3) U [4, 8)

    Interval sets over the reals are made of "RealInterval" objects,
    which is stated in the `interval_class` member.
    Also the builder() and utils() methods are defined in a standard way,
    which acts as reference for custom implementations as well.

    Note a reference to the class of "intervals making up these
    sets" is configured.

    Additionally, metadata for serializability are specified below:
    these two (class name and version) end up in the serializable dict
    to handle future 'schema changes', if there ever will be.
    """

    interval_class = RealInterval

    serializing_class = "RealIntervalSet"
    serializing_version = 1

    @staticmethod
    def builder():
        """
        Create an interval set builder.
        Other implementation need to simply replace `RealIntervalSet` here.
        """
        return IntervalGenericBuilder(
            interval_set_class=RealIntervalSet,
        )

    @staticmethod
    def utils():
        """
        Create an "utils" object, which offers standard intervalset* creation.
        Other implementation need to simply replace `RealIntervalSet` here.
        """
        return IntervalSetGenericUtils(
            interval_set_class=RealIntervalSet,
        )



"""
A utility to provide a "builder" interface for intervals, interval sets etc.
The actual class used to instantiate the interval* once completed is passed
when instantiating the builder: so, this "interface" can offer a standard
syntax yet result in the creation of intervals, interval sets and also
other types of intervalsets (e.g. enriched with metrics and so on).

Note: contrary to the 'utils' generic tools, this one is able to create
both intervals* and intervalsets* (depending on how initialized).
"""

from clothesline.algebra.symbols import PlusInf, MinusInf
from clothesline.interval_peg import IntervalPeg


class IntervalGenericBuilder:
    """
    A builder for some kind of intervals.
    An instance of this class supports a [] and () syntax
    for specifying the first peg, an action which creates another object
    that waits for a "completion" call with () or [] and eventually
    results in creation of an interval*.
    """

    class _builder:
        """
        An internal volatile class used to represent, internally,
        states such as `builder(first_peg)`, i.e. ready to accept
        another [] or () and complete the sequence by creating
        an instance of (the appropriate class of) interval*.
        """

        def __init__(self, finalizer, begin):
            """a _builder knows the first peg and which finalizer to use."""
            self._finalizer = finalizer
            self._begin = begin

        def _complete(self, value_end, included_end):
            """
            Common call used behind the () and [] calls used to set
            the second peg thus completing the building.
            """
            if value_end is Ellipsis:
                _value_end = PlusInf
                _included_end = False
            else:
                _value_end = value_end
                _included_end = included_end
            end_peg = IntervalPeg(_value_end, _included_end)
            return self._finalizer([self._begin, end_peg])

        def __call__(self, value):
            return self._complete(value, False)

        def __getitem__(self, value):
            return self._complete(value, True)

    def __init__(self, interval_class=None, interval_set_class=None):
        """
        When creating an IntervalGenericBuilder instance a class name
        for an interval class must be passed (a constructor ready to accept
        two pegs).
        Alternatively, to have the builder create intervalsets*, one passes
        the class for the corresponding intervalset*.
        Internally, suitable "finalizers" are crafted out of these inputs,
        so that the rest of the builder usage is the same for the two paths.
        """
        if interval_set_class is None:
            self._finalizer = lambda pegs: interval_class(*pegs)
        else:
            _interval_class = interval_set_class.interval_class
            self._finalizer = lambda pegs: interval_set_class(
                [_interval_class(*pegs)],
            )

    def _start_building(self, value, included):
        """
        Common call used behind the [] and () invocations, returns
        a _builder instance ready to accept the second peg of the interval.
        """
        if value is Ellipsis:
            _value = MinusInf
            _included = False
        else:
            _value = value
            _included = included
        return self._builder(self._finalizer, IntervalPeg(_value, _included))

    def __call__(self, value):
        """Initiate interval building from an excluded initial peg."""
        return self._start_building(value, included=False)

    def __getitem__(self, value):
        """Initiate interval building from an included initial peg."""
        return self._start_building(value, included=True)



"""
Specific exceptions for clothesline objects.
"""


class InvalidValueError(ValueError):
    """
    An attempt to build an object (e.g. IntervalPeg) with invalid
    or inconsistent values.
    """


class InvalidCombineEndState(ValueError):
    """
    An end state when closing the ''combine_intervals'
    with an unterminated open buffer to flush.
    """


class IndeterminateFormError(ValueError):
    """
    An indeterminate form arisen through operations
    involving symbols (e.g. "infinity - infinity")
    """


class MetricNotImplementedError(ValueError):
    """
    An interval*/intervalset* class is asked to do
    metric computations, but no metric is defined.
    """


class UnserializableItemError(ValueError):
    """
    An attempt to convert an interval*/set* to json is made
    for an object whose domain data type admits no serializable form.
    """


class UnparseableDictError(ValueError):
    """
    An attempt to parse an invalid dict back to an interval* or intervalset*
    """


class UnsupportedVersionDictError(ValueError):
    """
    The version of this dict is too new to be hydrated back to an object.
    """



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

setup(
    name='clothesline',
    version='0.1.0',
    author='Stefano Lottini',
    author_email='sl_@fastmail.com',
    package_dir={"": "src"},
    packages=find_packages(where='src'),
    # entry_points={
    #     "console_scripts": [
    #         "clothesline=clothesline:main",
    #     ],
    # },
    url='https://clothesline.readthedocs.io',
    license='LICENSE.txt',
    description='An algebra for intervals over a range of continuous values',
    long_description=(here / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    # install_requires=[
    #     "pytest",
    # ],
    python_requires=">=3.4.*",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        #
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="sets, intervals, algebra",
)



"""
A class containing the "interface" for creation ex-novo of intervalsets*,
with the key feature that the actual class to instantiate when creating these
is set upon creation of the IntervalSetGenericUtils instance.
In this way, a "utils" object can be spawned by IntervalSets or analogous
classes and with a common set of methods an "utils" creates intervalsets
of the appropriate type.
"""

from clothesline.exceptions import (
    UnparseableDictError,
    UnserializableItemError,
    UnsupportedVersionDictError,
)


class IntervalSetGenericUtils:
    """
    An "interval set utils" class. Instances are able to use the provided
    'name of an interval set class' and create standard out-of-the-box
    intervals.
    """

    def __init__(self, interval_set_class):
        """
        An instance of IntervalSetGenericUtils needs to know what class
        to use to create intervalsets* (and intervals*).
        From this knowledge, all other properties and instantiators
        are crafted internally (also to create the right type of objects).
        """
        interval_class = interval_set_class.interval_class
        self.set_instantiator = interval_set_class
        # the above would be: lambda intervals: interval_set_class(intervals)
        self.int_utils = interval_class.utils()
        self.serializing_class = interval_set_class.serializing_class
        self.serializing_version = interval_set_class.serializing_version

    def from_dict(self, input_dict):
        """
        Extract an instance of this interval set from a dict,
        using the provided serializability settings
        (including checking dict metadata match).
        """
        if self.serializing_class is None or self.serializing_version is None:
            raise UnserializableItemError
        #
        if input_dict.get("class") != self.serializing_class:
            raise UnparseableDictError
        # Here, in the future, version upgrade logic will be injected
        if input_dict.get("version", 0) > self.serializing_version:
            raise UnsupportedVersionDictError
        if input_dict.get("version") != self.serializing_version:
            raise UnparseableDictError
        #
        return self.set_instantiator(
            self.int_utils.from_dict(interval_dict)
            for interval_dict in input_dict["intervals"]
        )

    def empty(self):
        """
        Create the empty set.
        """
        return self.set_instantiator([])

    def open(self, value_begin, value_end):
        """
        Create an open interval set with finite boundaries.
        """
        return self.set_instantiator(
            [self.int_utils.open(value_begin, value_end)],
        )

    def closed(self, value_begin, value_end):
        """
        Create a closed interval set with finite boundaries.
        """
        return self.set_instantiator(
            [self.int_utils.closed(value_begin, value_end)],
        )

    def point(self, value):
        """
        Create a zero-length degenerate [x, x] point-line 'interval set'.
        """
        return self.set_instantiator([self.int_utils.point(value)])

    def low_slice(self, value_end, included=False):
        """
        Create an interval set from -inf to a certain value.
        """
        return self.set_instantiator(
            [self.int_utils.low_slice(value_end, included=included)],
        )

    def high_slice(self, value_begin, included=False):
        """
        Create an interval set from a value up to +inf.
        """
        return self.set_instantiator(
            [self.int_utils.high_slice(value_begin, included=included)],
        )

    def all(self):
        """
        Return the "whole of it" interval set.
        """
        return self.set_instantiator([self.int_utils.all()])

    def interval(self, value_begin, begin_included, value_end, end_included):
        """
        Directly create an interval set from the values
        and the open/closed specs.
        """
        return self.set_instantiator(
            [
                self.int_utils.interval(
                    value_begin,
                    begin_included,
                    value_end,
                    end_included,
                )
            ]
        )
